/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.constraint;

import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.createSimpleDescription;
import static jp.objectfanatics.commons.javassist.JavassistUtils.insertAfter;
import static jp.objectfanatics.commons.javassist.JavassistUtils.insertBefore;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAbstract;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAtomicInteger;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAtomicLong;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isBigDecimal;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isBigInteger;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isByte;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isDouble;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isFloat;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isInteger;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isLong;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitive;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveByte;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveDouble;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveFloat;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveInt;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveLong;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveNumber;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitiveShort;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isShort;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import javassist.bytecode.annotation.DoubleMemberValue;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalUseOfConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalUseOfParameterConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalUseOfReturnValueConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.impl.exception.IllegalUseOfParameterConstraintAnnotationExceptionImpl;
import jp.objectfanatics.assertion.weaver.impl.exception.IllegalUseOfReturnValueConstraintAnnotationExceptionImpl;
import jp.objectfanatics.commons.javassist.JavassistUtils;

// TODO: AssertionWeaver 系は全体的にかなりいじったので、全体を徹底的に整理しよう。 09.22.2009
/**
 * Abstract class for number comparing weavers.
 * 
 * @author Makoto Sato
 */
abstract public class AbstractNumberComparingAssertionWeaver extends AbstractAssertionWeaver {
	
	/**
	 * for asserting Integer and Long value is not null.
	 */
	private NotNullAssertionWeaver notNullAssertionWeaver = new NotNullAssertionWeaver();
	
	/**
	 * Comparator.
	 */
	private final String comparator;
	
	/**
	 * Comparative word for number out of bounds.
	 */
	private final String comparativeWordForNumberOutOfbounds;
	
	/**
	 * Constructor.
	 * 
	 * @param targetConstraintAnnotation
	 */
	public AbstractNumberComparingAssertionWeaver(Class<? extends java.lang.annotation.Annotation> targetConstraintAnnotation, String comparator, String comparativeWordForNumberOutOfbounds) {
		// TODO: assertion;
		super(targetConstraintAnnotation);
		this.comparator = comparator;
		this.comparativeWordForNumberOutOfbounds = comparativeWordForNumberOutOfbounds;
	}
	
	/**
	 * Returns annotation presentation string.
	 * 
	 * @return annotation presentation string.
	 */
	protected String getAnnotationPresentationString() {
		return targetConstraintAnnotation.getSimpleName();
	}
	
	@Override
	public void weaveParameterConstraintAssertion(CtBehavior ctBehavior, int paramIndex, Annotation annotation) throws IllegalUseOfParameterConstraintAnnotationException {
		
		CtClass paramType = JavassistUtils.getParameterType(ctBehavior, paramIndex);
		
		// weaving is valid only for byte, short, int, long, float, double, Byte, Short, Integer, Long, Float, Double, BigInteger, BigDecimal, AtomicInteger, and AtomicLong.
		if (!isValidType(paramType)) {
			String description = createSimpleDescription(ctBehavior, paramType, annotation);
			throw new IllegalUseOfParameterConstraintAnnotationExceptionImpl(description, paramIndex, ctBehavior, annotation);
		}
		
		try {
			// create assertionStatement
			String valueName = "parameter " + paramIndex + "(zero base)";
			String paramValueIdentifier = JavassistUtils.toParamValueIdentifier(ctBehavior, paramIndex);
			String assertionStatement = createAssertionStatement(ctBehavior, paramType, paramValueIdentifier, annotation, valueName, true);
			
			// weaves
			if (isAbstract(ctBehavior)) {
				// do nothing.
				return;
			}
			insertBefore(ctBehavior, assertionStatement);
			
			// delegates @NotNull assertion
			if (!isPrimitive(paramType)) notNullAssertionWeaver.weaveParameterConstraintAssertion(ctBehavior, paramIndex, annotation);
		} catch (InvalidAnnotationValueException e) {
			throw new IllegalUseOfParameterConstraintAnnotationExceptionImpl(e.getDescription(), paramIndex, ctBehavior, annotation);
		}
		
	}
	
	@Override
	public void weaveReturnValueConstraintAssertion(CtMethod ctMethod, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {
		
		CtClass returnType = JavassistUtils.getReturnType(ctMethod);
		
		// weaving is valid only for int, long, Integer and Long
		if (!isValidType(returnType)) {
			String description = createSimpleDescription(ctMethod, returnType, annotation);
			throw new IllegalUseOfReturnValueConstraintAnnotationExceptionImpl(description, ctMethod, annotation);
		}

		try {
			// delegates @NotNull assertion
			if (!isPrimitive(returnType)) notNullAssertionWeaver.weaveReturnValueConstraintAssertion(ctMethod, annotation);// TODO: Nullのweaverがabstractをはじいていることを確認。
			
			// weaves
			String valueName = "return value";
			String assertionStatement = createAssertionStatement(ctMethod, returnType, "$_", annotation, valueName, false);
			if (!isAbstract(ctMethod)) insertAfter(ctMethod, assertionStatement);
		} catch (InvalidAnnotationValueException e) {
			throw new IllegalUseOfReturnValueConstraintAnnotationExceptionImpl(e.getDescription(), ctMethod, annotation);
		}
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of byte.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @throws InvalidAnnotationValueException 
	 *             when min is out of the range of byte.
	 */
	protected void assertByte(CtBehavior ctBehavior, String valueName, double threshold) throws InvalidAnnotationValueException {
		if (threshold > Byte.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Byte.MAX_VALUE[=" + Byte.MAX_VALUE + "]");
		if (threshold < Byte.MIN_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "Byte.MIN_VALUE[=" + Byte.MIN_VALUE + "]");
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of short.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param min
	 * @throws InvalidAnnotationValueException 
	 *             when min is out of the range of short.
	 */
	protected void assertShort(CtBehavior ctBehavior, String valueName, double threshold) throws InvalidAnnotationValueException {
		if (threshold > Short.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Short.MAX_VALUE[=" + Short.MAX_VALUE + "]");
		if (threshold < Short.MIN_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "Short.MIN_VALUE[=" + Short.MIN_VALUE + "]");
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of int.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @throws InvalidAnnotationValueException 
	 *             when threshold is out of the range of int.
	 */
	protected void assertInt(CtBehavior ctBehavior, String valueName, double threshold) throws InvalidAnnotationValueException {
		if (threshold > Integer.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Integer.MAX_VALUE[=" + Integer.MAX_VALUE + "]");
		if (threshold < Integer.MIN_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "Integer.MIN_VALUE[=" + Integer.MIN_VALUE + "]");
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of long.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @throws InvalidAnnotationValueException 
	 *             when threshold is out of the range of long.
	 */
	protected void assertLong(CtBehavior ctBehavior, String valueName, double threshold) throws InvalidAnnotationValueException {
		if (threshold > Long.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Long.MAX_VALUE[=" + Long.MAX_VALUE + "]");
		if (threshold < Long.MIN_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "Long.MIN_VALUE[=" + Long.MIN_VALUE + "]");
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of float.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @throws InvalidAnnotationValueException 
	 *             when threshold is out of the range of float.
	 */
	protected void assertFloat(CtBehavior ctBehavior, String valueName, double threshold) throws InvalidAnnotationValueException {
		if (threshold > Float.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Float.MAX_VALUE[=" + Float.MAX_VALUE + "]");
		if (threshold < -Float.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "-Float.MAX_VALUE[=" + (-Float.MAX_VALUE) + "]");
	}
	
	/**
	 * Asserts that <code>threshold</code> is in the range of BigInteger.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @throws InvalidAnnotationValueException 
	 *             when threshold is out of the range of BigInteger.
	 */
	protected void assertBigInteger(CtBehavior ctBehavior, String valueName, double threshold) throws InvalidAnnotationValueException {
		if (threshold > Long.MAX_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, true, "Long.MAX_VALUE[=" + Long.MAX_VALUE + "]");
		if (threshold < Long.MIN_VALUE) throwExceptionForThresholdOutOfBounds(ctBehavior, valueName, threshold, false, "Long.MIN_VALUE[=" + Long.MIN_VALUE + "]");
	}
	
	/**
	 * Throws {@link IllegalUseOfConstraintAnnotationException} for threshold out of bounds.
	 * 
	 * @param ctBehavior
	 * @param valueName
	 * @param threshold
	 * @param isGreater
	 * @param target
	 * @throws InvalidAnnotationValueException 
	 */
	private void throwExceptionForThresholdOutOfBounds(CtBehavior ctBehavior, String valueName, double threshold, boolean isGreater, String target) throws InvalidAnnotationValueException {
		throw new InvalidAnnotationValueException("Specified value for " + valueName + " is " + ((isGreater ? "greater" : "less") + " than " + target) + " : @" + getAnnotationPresentationString() + "(" + threshold + ") : " + ctBehavior.getLongName());
	}
	
	private String createAssertionStatement(CtBehavior ctBehavior, CtClass type, String valueIdentifier, Annotation annotation, String valueName, boolean isParam) throws InvalidAnnotationValueException {
		double threshold = DoubleMemberValue.class.cast(annotation.getMemberValue("value")).getValue();
		if (Double.isNaN(threshold) || Double.isInfinite(threshold)) throw new InvalidAnnotationValueException("" + threshold + " is not supported by @" + getAnnotationPresentationString() + " at " + ctBehavior.getLongName());
		String assertionErrorConditionJavaSource;
		// byte
		if (isPrimitiveByte(type)) {
			assertByte(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " " + threshold;
		}
		// Byte
		else if (isByte(type)) {
			assertByte(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".byteValue() " + comparator + " " + threshold;
		}
		// short
		else if (isPrimitiveShort(type)) {
			assertShort(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " " + threshold;
		}
		// Short
		else if (isShort(type)) {
			assertShort(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".shortValue() " + comparator + " " + threshold;
		}
		// int
		else if (isPrimitiveInt(type)) {
			assertInt(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " " + threshold;
		}
		// Integer, AtomicInteger
		else if (isInteger(type) || isAtomicInteger(type)) {
			assertInt(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".intValue() " + comparator + " " + threshold;
		}
		// long
		else if (isPrimitiveLong(type)) {
			assertLong(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " (double)" + threshold;
		}
		// Long, AtomicLong
		else if (isLong(type) || isAtomicLong(type)) {
			assertLong(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".longValue() " + comparator + " (double)" + threshold;
		}
		// float
		else if (isPrimitiveFloat(type)) {
			assertFloat(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " (double)" + threshold;
		}
		// Float
		else if (isFloat(type)) {
			assertFloat(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".floatValue() " + comparator + " (double)" + threshold;
		}
		// double
		else if (isPrimitiveDouble(type)) {
			assertionErrorConditionJavaSource = valueIdentifier + " " + comparator + " (double)" + threshold;
		}
		// Double
		else if (isDouble(type)) {
			assertionErrorConditionJavaSource = valueIdentifier + ".doubleValue() " + comparator + " (double)" + threshold;
		}
		// BigInteger
		else if (isBigInteger(type)) {
			assertBigInteger(ctBehavior, valueName, threshold);
			assertionErrorConditionJavaSource = valueIdentifier + ".longValue() " + comparator + " (double)" + threshold;
		}
		// BigDecimal
		else if (isBigDecimal(type)) {
			assertionErrorConditionJavaSource = valueIdentifier + ".doubleValue() " + comparator + " (double)" + threshold;
		}
		// Unexpected type
		else {
			throw new IllegalStateException("Unexpected type : " + type.getName());
		}
		String message = valueName + "[=\"+" + valueIdentifier + "+\"] is " + comparativeWordForNumberOutOfbounds + " than " + threshold + "."; // TODO: use parameter name in the message.
		String assrtionStatement = isParam ? generateParameterAssrtionStatement(assertionErrorConditionJavaSource, message) : generateReturnValueAssrtionStatement(assertionErrorConditionJavaSource, message);
		return assrtionStatement;
	}
	
	// FIXME: isPrimitiveは boolean とか入るのでバグ。isPrimitiveNumberとか作成しよう。
	/**
	 * Returns true if, and only if, the specified type is byte, short, int, long, float, double, Byte, Short, Integer,
	 * Long, Float, Double, BigInteger, BigDecimal, AtomicInteger or AtomicLong.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is byte, short, int, long, float, double, Byte, Short, Integer,
	 *         Long, Float, Double, BigInteger, BigDecimal, AtomicInteger or AtomicLong.
	 */
	public static boolean isValidType(CtClass type) {
		return isPrimitiveNumber(type)  // primitive types
			|| isByte(type)             // java.lang.Byte
			|| isShort(type)            // java.lang.Short
			|| isInteger(type)          // java.lang.Integer
			|| isLong(type)             // java.lang.Long
			|| isFloat(type)            // java.lang.Float
			|| isDouble(type)           // java.lang.Double
			|| isBigInteger(type)       // java.math.BigInteger
			|| isBigDecimal(type)       // java.math.BigDecimal
			|| isAtomicInteger(type)    // java.util.concurrent.atomic.AtomicInteger
			|| isAtomicLong(type);      // java.util.concurrent.atomic.AtomicLong
	}
	
	/**
	 * Parameter か ReturnValue か認識できない状況で、値をつめるための例外。 TODO: 外部に公開するかどうかは要検討。
	 */
	class InvalidAnnotationValueException extends Exception {
		private final String description;
		
		public InvalidAnnotationValueException(String description) {
			super(description);
			this.description = description;
		}
		
		public String getDescription() {
			return description;
		}
	}
}
