/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.constraint;

import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.createSimpleDescription;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getReturnType;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAbstract;
import static jp.objectfanatics.commons.javassist.JavassistUtils.toParamValueIdentifier;
import javassist.CannotCompileException;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import jp.objectfanatics.assertion.constraints.NotNull;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalUseOfParameterConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.exception.IllegalUseOfReturnValueConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.impl.core.ConstraintAnnotationAssertionWeaver;
import jp.objectfanatics.assertion.weaver.impl.exception.IllegalUseOfParameterConstraintAnnotationExceptionImpl;
import jp.objectfanatics.assertion.weaver.impl.exception.IllegalUseOfReturnValueConstraintAnnotationExceptionImpl;
import jp.objectfanatics.commons.javassist.JavassistUtils;

/**
 * Implementation of {@link ConstraintAnnotationAssertionWeaver} for @{@link NotNull} annotation.
 * 
 * @author Makoto Sato
 */
public class NotNullAssertionWeaver extends AbstractAssertionWeaver {
	
	/**
	 * Constructor.
	 */
	public NotNullAssertionWeaver() {
		super(NotNull.class);
	}
	
	@Override
	public void weaveParameterConstraintAssertion(CtBehavior ctBehavior, int paramIndex, Annotation annotation) throws IllegalUseOfParameterConstraintAnnotationException {
		
		CtClass paramType = JavassistUtils.getParameterType(ctBehavior, paramIndex);

		
		// weaving is invalid for primitive type parameter
		if (paramType.isPrimitive()) {
			String description = createSimpleDescription(ctBehavior, paramType, annotation);
			throw new IllegalUseOfParameterConstraintAnnotationExceptionImpl(description, paramIndex, ctBehavior, annotation);
		}
		
		try {
			// weaves
			String paramValueIdentifier = toParamValueIdentifier(ctBehavior, paramIndex);
			String assertionErrorCondition = paramValueIdentifier + " == null";
			String message = "parameter " + paramIndex + "(zero base)" + " is null."; // TODO: use parameter name in the message.
			String assertionStatement = generateParameterAssrtionStatement(assertionErrorCondition, message);
			
			// weaves
			if (!isAbstract(ctBehavior)) ctBehavior.insertBefore(assertionStatement);
		}
		catch (CannotCompileException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
	
	@Override
	public void weaveReturnValueConstraintAssertion(CtMethod ctMethod, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {
		
		CtClass returnType = getReturnType(ctMethod);
		
		// weaving is invalid for primitive type parameter
		if (returnType.isPrimitive()) {
			String description = createSimpleDescription(ctMethod, returnType, annotation);
			throw new IllegalUseOfReturnValueConstraintAnnotationExceptionImpl(description, ctMethod, annotation);
		}
		try {
			// weaves
			String assertionErrorCondition = "$_ == null";
			String message = "return value is null.";
			String assertionStatement = generateReturnValueAssrtionStatement(assertionErrorCondition, message);
			if (!isAbstract(ctMethod)) ctMethod.insertAfter(assertionStatement);
		}
		catch (CannotCompileException e) {
			throw new IllegalStateException(e.getMessage(), e);
		}
	}
}
