/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.commons.javassist;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

import javassist.CannotCompileException;
import javassist.ClassPath;
import javassist.ClassPool;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtConstructor;
import javassist.CtMethod;
import javassist.Modifier;
import javassist.NotFoundException;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.AttributeInfo;
import javassist.bytecode.ClassFile;
import javassist.bytecode.CodeAttribute;
import javassist.bytecode.LineNumberAttribute;
import javassist.bytecode.LocalVariableAttribute;
import javassist.bytecode.MethodInfo;
import javassist.bytecode.ParameterAnnotationsAttribute;
import javassist.bytecode.annotation.Annotation;

// TODO: groupId=jp.objectfanatics.commons, artifactId=javassist として単独のアーティファクトに切り出しましょう。
public class JavassistUtils {
	
	public static CtClass[] getParameterTypesWithoutContextParameter(CtBehavior ctBehavior) {
		CtClass[] paramTypes = getParameterTypes(ctBehavior);
		if (paramTypes.length == 0) return paramTypes;
		if (hasContextParameter(ctBehavior)) {
			CtClass[] result = new CtClass[paramTypes.length-1];
			System.arraycopy(paramTypes, 1, result, 0, result.length);
			return result;
		} else { 
			return paramTypes;
		}
	}

	public static boolean hasContextParameter(CtBehavior ctBehavior) {
		return !isStatic(ctBehavior.getDeclaringClass()) && isInnerClass(ctBehavior.getDeclaringClass()) && CtConstructor.class.isInstance(ctBehavior);
	}
	
	public static boolean isInnerClass(CtClass ctClass) {
		return getDeclaringClass(ctClass) != null;
	}
	
	public static CtClass getDeclaringClass(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		try {
			return ctClass.getDeclaringClass();
		}
		catch (NotFoundException e) {
			// Mock test is impossible because CtMethod is final class.
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}


	
	/**
	 * Syntax sugar for {@link Modifier#isStatic(int)}
	 * 
	 * @param ctMember
	 *            target {@link CtBehavior}
	 * @return Returns true if, and only if, <code>ctBehavior</code> is static.
	 * @throws IllegalArgumentException
	 *             if <code>ctMember</code> is null
	 */
	public static boolean isStatic(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctMember is null.");
		return Modifier.isStatic(ctClass.getModifiers());
	}

	
	public static ClassPool getClassPool(CtBehavior ctBehavior) {
		return ctBehavior.getDeclaringClass().getClassPool();
	}
	
	public static CtClass getCtClass(ClassPool classPool, String fqcn) {
		if (classPool == null) throw new IllegalArgumentException("classPool is null.");
		if (fqcn == null) throw new IllegalArgumentException("fqcn is null.");
		try {
			return classPool.get(fqcn);
		}
		catch (NotFoundException e) {
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}
	
	public static CtClass getCtClass(ClassPool classPool, Annotation annotation) {
		if (classPool == null) throw new IllegalArgumentException("classPool is null.");
		if (annotation == null) throw new IllegalArgumentException("annotation is null.");
		return getCtClass(classPool, annotation.getTypeName());
	}
	
	public static String getSimpleName(ClassPool classPool, Annotation annotation) {
		if (classPool == null) throw new IllegalArgumentException("classPool is null.");
		if (annotation == null) throw new IllegalArgumentException("annotation is null.");
		return getCtClass(classPool, annotation.getTypeName()).getSimpleName();
	}
	
	// TODO: 以下テストケース作成。
	public static boolean isPrimitiveNumber(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return isPrimitiveByte(type)    // byte
			|| isPrimitiveShort(type)   // short
			|| isPrimitiveInt(type)     // int
			|| isPrimitiveLong(type)    // long
			|| isPrimitiveFloat(type)   // float
			|| isPrimitiveDouble(type); // double
	}
	
	public static CtClass getParameterType(CtBehavior ctBehavior, int parameterIndex) {
		return getParameterTypesWithoutContextParameter(ctBehavior)[parameterIndex];
	}
	
	public static boolean isConstructor(CtBehavior ctBehavior) {
		if (CtConstructor.class.isInstance(ctBehavior)) return true;
		if (CtMethod.class.isInstance(ctBehavior)) return false;
		throw new IllegalStateException("Unexpected CtBehavior : " + ctBehavior.getClass());
	}
	
	public static String getPackageName(CtBehavior ctBehavior) {
		return ctBehavior.getDeclaringClass().getPackageName();
	}
	
	public static void insertBefore(CtBehavior ctBehavior, String statement) {
		try {
			ctBehavior.insertBefore(statement);
		}
		catch (CannotCompileException e) {
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}
	
	public static void insertAfter(CtBehavior ctBehavior, String statement) {
		try {
			ctBehavior.insertAfter(statement);
		}
		catch (CannotCompileException e) {
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}
	
	/**
	 * Returns a handle for dealing with the target class file.
	 * 
	 * @param classfileBuffer
	 *            class file bytes
	 * @return a handle for dealing with the target class file
	 * @throws  
	 * @throws IOException if classfileBuffer is not class-file-bytes
	 */
	public static CtClass getCtClass(ClassPool classPool, byte[] classfileBuffer) throws IOException {
		InputStream in = new ByteArrayInputStream(classfileBuffer);
		CtClass ctClass = classPool.makeClass(in);
		return ctClass;
	}
	
	
	
	// てすとちう

	
	
	
	
	
	
	// --------------------------------------------------------------------------------------------
	// Unclassified
	// --------------------------------------------------------------------------------------------
	// Checked(9.19.2009), Tested(9.19.2009)
	/**
	 * Returns true if the specified ctClass is annotated with the annotation expressed by <code>annotationFqcn</code>.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @param annotationFqcn
	 *            fqcn of target {@link Annotation}
	 * @return true if the specified ctClass is annotated with the annotation expressed by <code>annotationFqcn</code>
	 */
	public static boolean isAnnotated(CtClass ctClass, String annotationFqcn) {
		try {
			getAnnotation(ctClass, annotationFqcn);
			return true;
		}
		catch (AnnotationNotFoundException e) {
			return false;
		}
	}
	
	// Checked(9.19.2009), Tested(9.19.2009)
	/**
	 * Returns an {@link Annotation} annotating ctClass.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @param annotationFqcn
	 *            fqcn of target {@link Annotation}
	 * @return target {@link Annotation}
	 * @throws AnnotationNotFoundException
	 *             if target annotation is not found.
	 */
	public static Annotation getAnnotation(CtClass ctClass, String annotationFqcn) throws AnnotationNotFoundException {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		if (annotationFqcn == null) throw new IllegalArgumentException("annotationFqcn is null.");
		try {
			AnnotationsAttribute attr = getAnnotationsAttribute(ctClass);
			Annotation annotation = attr.getAnnotation(annotationFqcn);
			if (annotation != null) return annotation;
		}
		catch (AnnotationsAttributeNotFoundException e) {
			// do nothing.
		}
		throw new AnnotationNotFoundException("ctClass = " + ctClass.getName() + ", annotationFqcn = " + annotationFqcn);
	}
	/**
	 * Thrown if annotation is not found.
	 */
	public static class AnnotationNotFoundException extends Exception {
		public AnnotationNotFoundException(String message) { super(message); }
	}
	
	// TODO: クラスやインタフェース互換性関連のユーティリティメソッドは、全体的に整合性を考えて、名前を十分に検討し、整理しましょう。
	/**
	 * A syntax sugar for getAssignableTypesExceptItself(ctBehavior.getDeclaringClass())
	 * 
	 * @param ctMethod
	 *            {@link CtMethod}
	 * @return assignable types (not include itself)
	 */
	public static CtClass[] getAssignableTypesExceptItself(CtBehavior ctBehavior) {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		CtClass declaringClass = ctBehavior.getDeclaringClass();
		return getAssignableTypesExceptItself(declaringClass);
	}
	
	// Checked(9.19.2009), Tested(9.19.2009)
	/**
	 * Returns assignable types (not include itself).
	 * 
	 * @param ctClass
	 *            {@link CtClass}
	 * @return assignable types (not include itself)
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 */
	public static CtClass[] getAssignableTypesExceptItself(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		Set<CtClass> assignableSuperTypes = new LinkedHashSet<CtClass>();
		addAll(assignableSuperTypes, getAncestorClasses(ctClass));
		addAll(assignableSuperTypes, getAssignableInterfacesExceptItself(ctClass));
		return assignableSuperTypes.toArray(new CtClass[0]);
	}
	
	// TODO: クラスやインタフェース互換性関連のユーティリティメソッドは、全体的に整合性を考えて、名前を十分に検討し、整理しましょう。
	// Checked(9.19.2009), Tested(9.19.2009)
	/**
	 * Returns ancestor classes (not include itself).
	 * <p/>
	 * Interface's super class is java.lang.Object.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return ancestor classes (not include itself)
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 */
	public static CtClass[] getAncestorClasses(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		Set<CtClass> ancestorClassess = new LinkedHashSet<CtClass>();
		CtClass current = ctClass;
		int count = 0;
		final int loopLimit = 10000; // to avoid infinite loop. it's just a safeguard.
		while (true) {
			CtClass superClass = getSuperClass(current);
			if (superClass == null) return ancestorClassess.toArray(new CtClass[0]);
			ancestorClassess.add(superClass);
			current = superClass;
			if (count++ > loopLimit) throw new IllegalStateException("loop counter has reached at " + loopLimit + ". ctClass = " + ctClass.getName() + ". It may be a bug. Please send a bug report.");
		}
	}
	
	// Checked(9.19.2009), Tested(9.19.2009)
	/**
	 * Returns new {@link ClassPool} created with system path.
	 * 
	 * @return new {@link ClassPool} created with system path
	 */
	public static ClassPool createClassPoolWithSystemPath() {
		ClassPool classPool = new ClassPool();
		classPool.appendSystemPath();
		return classPool;
	}
	
	// Checked(9.19.2009), Tested(9.19.2009)
	/**
	 * Creates class pool with system classpaths and extra classpaths.
	 * 
	 * @param extraClasspaths
	 *            extra classpaths
	 * @return a class pool configured with system classpaths and extra classpaths
	 */
	public static ClassPool createClassPool(String[] extraClasspaths) {
		if (extraClasspaths == null) throw new IllegalArgumentException("extraClasspaths is null.");
		ClassPool classPool = createClassPoolWithSystemPath();
		for (int i = 0; i < extraClasspaths.length; i++) {
			String classpath = extraClasspaths[i];
			if (classpath == null) throw new IllegalArgumentException("extraClasspaths[" + i + "] is null.");
			appendClassPath(classPool, classpath);
		}
		return classPool;
	}
	
	// Checked(9.19.2009), Tested(9.19.2009)
	/**
	 * Returns annotations.
	 * 
	 * @param ctBehavior
	 *            target {@link CtBehavior}
	 * @param target
	 *            parameter index
	 * @return annotations
	 * @throws ArrayIndexOutOfBoundsException
	 *             if paramIndex is out of range
	 */
	public static Annotation[] getParameterAnnotations(CtBehavior ctBehavior, int paramIndex) {
		return getParametersAnnotations(ctBehavior)[paramIndex];
	}
	
	// Checked(9.19.2009), Tested(9.19.2009)
	/**
	 * Returns all parameter's annotations.
	 * 
	 * @param ctBehavior
	 *            {@link CtBehavior}
	 * @return all parameter's annotations
	 */
	public static Annotation[][] getParametersAnnotations(CtBehavior ctBehavior) {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		try {
			return getParameterAnnotationsAttribute(ctBehavior).getAnnotations();
		}
		catch (ParameterAnnotationsAttributeNotFoundException e) {
			return new Annotation[getParameterTypesWithoutContextParameter(ctBehavior).length][0];
		}
	}
	
	// Checked(9.19.2009), Tested(9.19.2009)
	/**
	 * Returns {@link Annotation}[] annotating for <code>ctBehavior</code>.
	 * 
	 * @param ctBehavior
	 *            {@link CtBehavior}
	 * @return {@link Annotation}[] annotating for <code>ctBehavior</code>
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null.
	 */
	public static Annotation[] getAnnotations(CtBehavior ctBehavior) {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		try {
			AnnotationsAttribute attr = getAnnotationsAttribute(ctBehavior);
			return attr.getAnnotations();
		}
		catch (AnnotationsAttributeNotFoundException e) {
			return new Annotation[0];
		}
	}
	
	
	/**
	 * Returns {@link AnnotationsAttribute} annotating for the ctMethod.
	 * 
	 * @param ctBehavior
	 *            {@link CtMethod}
	 * @return {@link AnnotationsAttribute} annotating for the ctMethod
	 * @throws AnnotationsAttributeNotFoundException
	 *             if ParameterAnnotationsAttribute not found
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null.
	 */
	static AnnotationsAttribute getAnnotationsAttribute(CtBehavior ctBehavior) throws AnnotationsAttributeNotFoundException {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		MethodInfo methodInfo = ctBehavior.getMethodInfo();
		if (methodInfo == null) throw new IllegalStateException("methodInfo is not supposed to be null. please check the specs of javassist and java VM.");
		AttributeInfo annotationsAttribute = methodInfo.getAttribute(AnnotationsAttribute.visibleTag);
		if (annotationsAttribute == null) throw AnnotationsAttributeNotFoundException.SINGLETON;
		return (AnnotationsAttribute) annotationsAttribute;
	}
	
	// TODO: jp.objectfanatics.commons.CollectionUtils とかに移動を検討しましょう。
	// Checked(9.17.2009), Tested(9.17.2009)
	/**
	 * Adds all elements of <code>source</code> into <code>destination</code>.
	 * 
	 * @param <T>
	 *            parameterized type
	 * @param destination
	 *            destination
	 * @param source
	 *            source elements
	 * @throws IllegalArgumentException
	 *             if destination is null
	 * @throws IllegalArgumentException
	 *             if source is null
	 */
	public static <T> void addAll(Collection<T> destination, T[] source) {
		if (destination == null) throw new IllegalArgumentException("destination is null.");
		if (source == null) throw new IllegalArgumentException("source is null.");
		for (T element : source) destination.add(element);
	}
	
	// Checked(9.17.2009), Tested(9.17.2009)
	/**
	 * Returns {@link ParameterAnnotationsAttribute} of the specified {@link CtBehavior}.
	 * 
	 * @param ctBehavior
	 *            {@link CtBehavior}
	 * @return {@link ParameterAnnotationsAttribute} of the specified {@link CtBehavior}
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null
	 * @throws ParameterAnnotationsAttributeNotFoundException
	 *             if ParameterAnnotationsAttribute not found
	 */
	public static ParameterAnnotationsAttribute getParameterAnnotationsAttribute(CtBehavior ctBehavior) throws ParameterAnnotationsAttributeNotFoundException {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		MethodInfo methodInfo = ctBehavior.getMethodInfo();
		if (methodInfo == null) throw new IllegalStateException("methodInfo is not supposed to be null. please check the specs of javassist and java VM.");
		ParameterAnnotationsAttribute parameterAnnotationAttribute = (ParameterAnnotationsAttribute) methodInfo.getAttribute(ParameterAnnotationsAttribute.visibleTag);
		if (parameterAnnotationAttribute == null) {
			throw ParameterAnnotationsAttributeNotFoundException.SINGLETON;
		}
		else {
			return parameterAnnotationAttribute;
		}
	}
	public static class ParameterAnnotationsAttributeNotFoundException extends Exception {
		public static final ParameterAnnotationsAttributeNotFoundException SINGLETON = new ParameterAnnotationsAttributeNotFoundException();
		private ParameterAnnotationsAttributeNotFoundException() { super("AnnotationsAttribute not found."); }
	}
	
	// Checked(9.17.2009), Tested(9.17.2009)
	/**
	 * Returns true if <code>annotation</code> is annotated by the annotation the fqcn of which is <code>metaAnnotationFqcn</code>.
	 * 
	 * @param classPool
	 *            {@link ClassPool}
	 * @param annotation
	 *            {@link Annotation}
	 * @param metaAnnotationFqcn
	 *            fqcn of meta annotation class
	 * @return true if <code>annotation</code> is annotated by the annotation the fqcn of which is <code>metaAnnotationFqcn</code>
	 * @throws IllegalArgumentException
	 *             if <code>classPool</code> is null
	 * @throws IllegalArgumentException
	 *             if <code>annotation</code> is null
	 * @throws IllegalArgumentException
	 *             if <code>metaAnnotationFqcn</code> is null
	 */
	public static boolean isAnnotatedWithMetaAnnotation(ClassPool classPool, Annotation annotation, String metaAnnotationFqcn) {
		if (classPool == null) throw new IllegalArgumentException("classPool is null.");
		if (annotation == null) throw new IllegalArgumentException("annotation is null.");
		if (metaAnnotationFqcn == null) throw new IllegalArgumentException("metaAnnotationFqcn is null.");
		AnnotationsAttribute annotationAttribute = getAnnotationsAttribute(classPool, annotation);
		Annotation metaAnnotation = annotationAttribute.getAnnotation(metaAnnotationFqcn);
		return metaAnnotation != null;
	}
	
	// Checked(9.17.2009), Tested(9.17.2009)
	/**
	 * Returns {@link AnnotationsAttribute} annotating for the <code>annotation</code>.
	 * 
	 * @param classPool
	 *            {@link ClassPool}
	 * @param annotation
	 *            {@link Annotation}
	 * @throws IllegalArgumentException
	 *             if <code>classPool</code> is null
	 * @throws IllegalArgumentException
	 *             if <code>annotation</code> is null
	 * @return {@link AnnotationsAttribute} annotating for the <code>annotation</code>
	 * @throws AnnotationsAttributeNotFoundException
	 *             if AnnotationsAttribute not founds(ex. no annotation)
	 */
	public static AnnotationsAttribute getAnnotationsAttribute(ClassPool classPool, Annotation annotation) {
		try {
			if (classPool == null) throw new IllegalArgumentException("classPool is null.");
			if (annotation == null) throw new IllegalArgumentException("annotation is null.");
			CtClass annotationCtClass = classPool.get(annotation.getTypeName());
			return getAnnotationsAttribute(annotationCtClass);
		}
		catch (NotFoundException e) {
			// TODO: IllegalStateException 系の整理。
			throw new IllegalStateException("Class " + annotation.getTypeName() + "not found. Please check your classpath", e);
		}
		catch (AnnotationsAttributeNotFoundException e) {
			// parameter "annotation"'s scope is runtime, so it always has its AnnotationAttribute.
			throw new IllegalStateException("It's bug. Please report.");
		}
	}
	
	// Checked(9.17.2009), Tested(9.17.2009)
	/**
	 * Returns {@link AnnotationsAttribute} annotating for the <code>ctClass</code>.
	 * 
	 * @param ctClass
	 *            {@link CtClass}
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 * @return {@link AnnotationsAttribute} annotating for the <code>ctClass</code>
	 * @throws AnnotationsAttributeNotFoundException if AnnotationsAttribute not founds(ex. no annotation)
	 */
	public static AnnotationsAttribute getAnnotationsAttribute(CtClass ctClass) throws AnnotationsAttributeNotFoundException {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		ClassFile classFile = ctClass.getClassFile();
		if (classFile == null) throw new IllegalStateException("Can't get classFile.");
		AttributeInfo annotationsAttribute = classFile.getAttribute(AnnotationsAttribute.visibleTag);
		if (annotationsAttribute == null) throw AnnotationsAttributeNotFoundException.SINGLETON;
		return (AnnotationsAttribute) annotationsAttribute;
	}
	public static class AnnotationsAttributeNotFoundException extends Exception {
		public static final AnnotationsAttributeNotFoundException SINGLETON = new AnnotationsAttributeNotFoundException();
		private AnnotationsAttributeNotFoundException() { super("AnnotationsAttribute not found."); }
	}
	
	// Checked(9.17.2009), Tested(9.17.2009)
	/**
	 * Returns declared behaviors, without same methods with different return types.
	 * 
	 * @see #getUniqueDeclaredMethods(CtClass)
	 * @param ctClass
	 *            {@link CtClass}
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 * @return declared behaviors, without same methods with different return types
	 */
	public static CtBehavior[] getUniqueDeclaredBehaviors(CtClass ctClass) {
		List<CtBehavior> userDeclaredBehaviors = new ArrayList<CtBehavior>();
		// methods
		for (CtMethod ctMethod : getUniqueDeclaredMethods(ctClass)) {
			userDeclaredBehaviors.add(ctMethod);
		}
		// constructors
		for (CtBehavior declaredConstructor : ctClass.getDeclaredConstructors()) {
			userDeclaredBehaviors.add(declaredConstructor);
		}
		return userDeclaredBehaviors.toArray(new CtBehavior[userDeclaredBehaviors.size()]);
	}
	
	// Checked(9.17.2009), Tested(9.17.2009)
	/**
	 * Returns declared methods, without same methods with different return types.
	 * <p/>
	 * {@link CtClass#getDeclaredMethods()} returns multiple methods when an overriding method has a different return type.
	 * 
	 * @param ctClass
	 *            {@link CtClass}
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 * @return declared methods, without same methods with different return types
	 */
	public static CtMethod[] getUniqueDeclaredMethods(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		
		// buffer for return value
		List<CtMethod> result = new ArrayList<CtMethod>();
		
		// note : ctClass.getDeclaredMethods() returns multiple methods when a super method is overridden with different return type  
		CtMethod[] ambiguousDeclaredMethods = ctClass.getDeclaredMethods();
		
		for (CtMethod ambiguousDeclaredMethod : ambiguousDeclaredMethods) {
			String methodName = ambiguousDeclaredMethod.getName();
			CtClass[] parameterTypes = getParameterTypesWithoutContextParameter(ambiguousDeclaredMethod);
			
			String ambiguousReturnTypeName = getReturnType(ambiguousDeclaredMethod).getName();
			
			CtMethod uniqueDeclaredMethod = getDeclaredMethod(ctClass, methodName, parameterTypes);
			String uniqueReturnTypeName = getReturnType(uniqueDeclaredMethod).getName();
			
			if (ambiguousReturnTypeName.equals(uniqueReturnTypeName)) {
				result.add(ambiguousDeclaredMethod);
			}
			else {
				// do nothing. same method with different return type.
			}
		}
		return result.toArray(new CtMethod[result.size()]);
	}
	
	// Checked(9.17.2009), Tested(9.17.2009)
	/**
	 * Returns parameter names.
	 * 
	 * @param ctBehavior
	 *            target {@link CtBehavior}
	 * @return parameter names
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null
	 * @throws NoAvailableParameterNameInfoException
	 *             if there is no parameter name information
	 */
	public static String[] getParameterNames(CtBehavior ctBehavior) throws NoAvailableParameterNameInfoException {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		
		// no parameter
		if (getParameterTypesWithoutContextParameter(ctBehavior).length == 0) return new String[0];
		
		// a variable for return value
		String[] parameterNames = new String[getParameterTypesWithoutContextParameter(ctBehavior).length];
		
		// getting code attribute
		CodeAttribute codeAttribute = ctBehavior.getMethodInfo().getCodeAttribute();
		if (codeAttribute == null) {
			// native or abstract TODO: add URL for the reason of this.
			throw NoAvailableParameterNameInfoException.CODE_ATTRIBUTE_NOT_FOUND;
		}
		
		
		// getting local variable table
		LocalVariableAttribute locals = (LocalVariableAttribute) codeAttribute.getAttribute("LocalVariableTable");
		if (locals == null) {
			// TODO: add URL below
			// LocalVariableTable attribute is disabled. (LocalVariableTable attribute is an optional variable-length attribute of a Code attribute. see VM Spec.)
			throw NoAvailableParameterNameInfoException.LOCAL_VARIABLE_ATTRIBUTE_NOT_FOUND;
		}
		
// for debug. delete this after stabilized.
//		for (int i = 0; i < locals.tableLength(); i++) {
//			System.out.println(i + ", start_pc = " + locals.startPc(i) + ", index = " + locals.index(i) + " : " + locals.variableName(i));
//		}
		
		// checking "this"
		boolean isThisExist = false;
		for (int i = 0; i < locals.tableLength(); i++) {
			if (locals.startPc(i) != 0) continue;
			if ("this".equals(locals.variableName(i))) {
				isThisExist = true;
				break;
			}
		}
		
		for (int i = 0; i < locals.tableLength(); i++) {
			if (locals.startPc(i) != 0) continue; // skip 
			// instance function
			//   index == 0 -> this
			//   index == 1 -> first parameter
			//   index == 2 -> second parameter
			// static function
			//   index == 0 -> first parameter
			//   index == 1 -> second parameter
			int index = locals.index(i);
			int paramBeginIndex = isThisExist ? 1 : 0;
			int paramEndIndex = paramBeginIndex + parameterNames.length - 1;
			boolean isParameter = (index >= paramBeginIndex) && (index <= paramEndIndex);
			if (isParameter) {
				parameterNames[index - paramBeginIndex] = locals.variableName(i);
// for debug. delete this after stabilized.
//				System.out.println("param["+(index - paramBeginIndex)+"] = " + (parameterNames[index - paramBeginIndex]));
			}
		}
		return parameterNames;
	}
	// TODO: javadoc
	public static class NoAvailableParameterNameInfoException extends Exception {
		public static final NoAvailableParameterNameInfoException CODE_ATTRIBUTE_NOT_FOUND = new NoAvailableParameterNameInfoException("Code attribute not found.");
		public static final NoAvailableParameterNameInfoException LOCAL_VARIABLE_ATTRIBUTE_NOT_FOUND = new NoAvailableParameterNameInfoException("LocalVariable attribute not found.");
		private NoAvailableParameterNameInfoException(String message) { super(message); }
	}
	
	// Checked(9.16.2009), Tested(9.16.2009)
	/**
	 * Returns line number of the specified {@link CtBehavior}, or -1 if line number is not available.
	 * 
	 * @param ctBehavior
	 *            target {@link CtBehavior}
	 * @return line number of the specified {@link CtBehavior}, or -1 if line number is not available
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null
	 * @throws LineNumberNotAvailableException
	 */
	public static int getLineNumberEvenIfAvailableLineNumberInfoNotAvailable(CtBehavior ctBehavior) {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		try {
			return getLineNumber(ctBehavior);
		}
		catch (LineNumberNotAvailableException e) {
			return -1;
		}
	}
	
	// Checked(9.14.2009), Tested(9.14.2009)
	/**
	 * Returns line number of the specified {@link CtBehavior}.
	 * 
	 * @param ctBehavior
	 *            target {@link CtBehavior}
	 * @return line number of the specified {@link CtBehavior}
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null
	 * @throws LineNumberNotAvailableException
	 *             if line number is not available
	 */
	public static int getLineNumber(CtBehavior ctBehavior) throws LineNumberNotAvailableException {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		CodeAttribute codeAttribute = ctBehavior.getMethodInfo().getCodeAttribute();
		if (codeAttribute == null) throw LineNumberNotAvailableException.SINGLETON;
		LineNumberAttribute attr = (LineNumberAttribute) codeAttribute.getAttribute(LineNumberAttribute.tag);
		if (attr == null) throw LineNumberNotAvailableException.SINGLETON;  // TODO: test
		return attr.lineNumber(0);
	}
	public static class LineNumberNotAvailableException extends Exception {
		public static final LineNumberNotAvailableException SINGLETON = new LineNumberNotAvailableException();
		private LineNumberNotAvailableException() { super("Line number is not available."); }
	}
	
	// Checked(9.14.2009), Tested(9.14.2009)
	/**
	 * Returns a parameter value identifier for the specified parameter index. It is used in the source code to be compiled by javassist.
	 * 
	 * @param paramIndex
	 *            target parameter index
	 * @throws IllegalArgumentException
	 *             if <code>paramIndex</code> is less than zero
	 * @return a parameter value identifier for the specified parameter index.
	 */
	public static String toParamValueIdentifier(CtBehavior ctBehavior, int paramIndex) {
		if (hasContextParameter(ctBehavior)) paramIndex++;
		if (paramIndex < 0) throw new IllegalArgumentException("paramIndex is less than zero.");
		return "$" + (paramIndex + 1);
	}
	
	
	// --------------------------------------------------------------------------------------------
	// Syntax sugars
	// --------------------------------------------------------------------------------------------
	// Checked(9.14.2009), Tested(9.14.2009)
	/**
	 * Returns source file name of <code>ctBehavior</code>'s declaring class.
	 * 
	 * @param ctBehavior
	 *            target {@link CtBehavior}
	 * @return source file name
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null
	 */
	public static String getSourceFileName(CtBehavior ctBehavior) {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		return getSourceFileName(ctBehavior.getDeclaringClass());
	}
	
	// Checked(9.14.2009), Tested(9.14.2009)
	/**
	 * Returns source file name.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return source file name
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 */
	public static String getSourceFileName(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		String sourceFileName = ctClass.getClassFile().getSourceFile();
		// TODO: 例外作成。全体的に、RuntimeNotFoundExceptionとか作成する？
		if (sourceFileName == null) throw new IllegalStateException("sourceFileName not found.");
		return sourceFileName;
	}
	
	// Checked(9.14.2009), Tested(9.14.2009)
	/**
	 * Returns true if, and only if, the type of ctClass is void.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return true if, and only if, return type is void
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 */
	public static boolean isVoid(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		return ctClass.isPrimitive() && "void".equals(ctClass.getName());
	}
	
	// Checked(9.14.2009), Tested(9.14.2009)
	/**
	 * Syntax sugar for {@link Modifier#isPrivate(int)}
	 * 
	 * @param ctBehavior
	 *            target {@link CtBehavior}
	 * @return Returns true if, and only if, <code>ctBehavior</code> is private.
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null
	 */
	public static boolean isPrivate(CtBehavior ctBehavior) {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		return Modifier.isPrivate(ctBehavior.getModifiers());
	}
	// Checked(9.14.2009), Tested(9.14.2009)
	/**
	 * Syntax sugar for {@link Modifier#isAbstract(int)}
	 * 
	 * @param ctBehavior
	 *            target {@link CtBehavior}
	 * @return Returns true if, and only if, <code>ctBehavior</code> is abstract.
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null
	 */
	public static boolean isAbstract(CtBehavior ctBehavior) {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		return Modifier.isAbstract(ctBehavior.getModifiers());
	}
	
	
	// --------------------------------------------------------------------------------------------
	// Type testers
	// --------------------------------------------------------------------------------------------
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is primitive boolean.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is primitive boolean.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isPrimitiveBoolean(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return boolean.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is primitive char.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is primitive char.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isPrimitiveChar(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return char.class.getName().equals(type.getName());
	}

	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is primitive byte.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is primitive byte.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isPrimitiveByte(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return byte.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is short.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is short.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isPrimitiveShort(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return short.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is int.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is int.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isPrimitiveInt(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return int.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is long.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is long.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isPrimitiveLong(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return long.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is float.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is float.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isPrimitiveFloat(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return float.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is double.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is double.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isPrimitiveDouble(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return double.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Boolean.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Boolean.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isBoolean(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Boolean.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Character.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Character.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isCharacter(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Character.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Byte.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Byte.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isByte(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Byte.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Short.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Short.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isShort(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Short.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Integer.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Integer.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isInteger(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Integer.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is AtomicInteger.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is AtomicInteger.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isAtomicInteger(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return AtomicInteger.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Long.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Long.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isLong(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Long.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is AtomicLong.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is AtomicLong.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isAtomicLong(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return AtomicLong.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Float.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Float.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isFloat(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Float.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is Double.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is Double.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isDouble(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return Double.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is BigInteger.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is BigInteger.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isBigInteger(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return BigInteger.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is BigDecimal.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is BigDecimal.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isBigDecimal(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return BigDecimal.class.getName().equals(type.getName());
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is compatible with CharSequence.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is compatible with CharSequence.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isAssignableToCharSequence(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return isSelfOrImplementor(type, CharSequence.class);
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is compatible with java.util.Collection.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is compatible with java.util.Collection.
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isAssignableToCollection(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return isSelfOrImplementor(type, Collection.class);
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is primitive.
	 * 
	 * @param type
	 *            target type
	 * @return true if, and only if, the specified type is primitive
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 */
	public static boolean isPrimitive(CtClass type) {
		if (type == null) throw new IllegalArgumentException("type is null.");
		return isPrimitiveBoolean(type) // boolean
			|| isPrimitiveChar(type)    // char
			|| isPrimitiveByte(type)    // byte
			|| isPrimitiveShort(type)   // short
			|| isPrimitiveInt(type)     // int
			|| isPrimitiveLong(type)    // long
			|| isPrimitiveFloat(type)   // float
			|| isPrimitiveDouble(type); // double
	}
	
	// TODO: クラスやインタフェース互換性関連のユーティリティメソッドは、全体的に整合性を考えて、名前を十分に検討し、整理しましょう。
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is compatible with the interface type expressed by <code>interfaceTypeFqcn</code>.
	 * 
	 * @param type
	 *            target type
	 * @param interfaceTypeFqcn
	 *            FQCN of interface type
	 * @return true if, and only if, the specified type is compatible with the interface type expressed by <code>interfaceTypeFqcn</code>
	 * @throws IllegalArgumentException
	 *             if <code>type</code> is null
	 * @throws IllegalArgumentException
	 *             if <code>interfaceTypeFqcn</code> is null
	 */
	public static boolean isSelfOrImplementor(CtClass type, String interfaceTypeFqcn) {
		// null check
		if (type == null) throw new IllegalArgumentException("type is null.");
		if (interfaceTypeFqcn == null) throw new IllegalArgumentException("interfaceTypeFqcn is null.");
		// self
		if (type.isInterface() && interfaceTypeFqcn.equals(type.getName())) return true;
		// for implementor
		for (CtClass implementedInterface : getAssignableInterfacesExceptItself(type)) {
			if (interfaceTypeFqcn.equals(implementedInterface.getName())) return true;
		}
		// no match
		return false;
	}
	
	// TODO: クラスやインタフェース互換性関連のユーティリティメソッドは、全体的に整合性を考えて、名前を十分に検討し、整理しましょう。
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns all assignable interfaces except for itself.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return all assignable interfaces except for itself
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 * @throws CheckedExceptionSuppressorException
	 *             if NotFoundException is thrown
	 */
	public static CtClass[] getAssignableInterfacesExceptItself(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		Set<CtClass> result = new HashSet<CtClass>();
		for (CtClass superInterface : getInterfaces(ctClass)) {
			result.add(superInterface);
			addAll(result, getAssignableInterfacesExceptItself(superInterface));
		}
		for (CtClass ancestorClass : getAncestorClasses(ctClass)) {
			for (CtClass superInterface : getInterfaces(ancestorClass)) {
				result.add(superInterface);
				addAll(result, getAssignableInterfacesExceptItself(superInterface));
			}
		}
		return result.toArray(new CtClass[result.size()]);
	}
	
	// TODO: クラスやインタフェース互換性関連のユーティリティメソッドは、全体的に整合性を考えて、名前を十分に検討し、整理しましょう。
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Returns true if, and only if, the specified type is assignable to the interface type expressed by <code>interfaceTypeFqcn</code>.
	 * 
	 * @param type
	 *            target type
	 * @param clazz
	 *            interface type
	 * @return true if, and only if, the specified type is assignable to the interface type expressed by <code>interfaceTypeFqcn</code>
	 */
	public static boolean isSelfOrImplementor(CtClass type, Class<?> interfaceType) {
		// null check
		if (type == null) throw new IllegalArgumentException("type is null.");
		if (interfaceType == null) throw new IllegalArgumentException("interfaceType is null.");
		// non-interface
		if (!interfaceType.isInterface()) throw new IllegalArgumentException("interfaceType is not interface : " + interfaceType.getName());
		// delegate
		return isSelfOrImplementor(type, interfaceType.getName());
	}
	
	// TODO: checked exception suppressor 系は、 CheckedExceptionSuppressor を作成して、そちらに移動したほうがよい。
	// --------------------------------------------------------------------------------------------
	// Checked exception suppressors
	// --------------------------------------------------------------------------------------------
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link CtClass#toBytecode()}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link CheckedExceptionSuppressorException}.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return return value of {@link CtClass#toBytecode()}
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null.
	 * @throws CheckedExceptionSuppressorException
	 *             if IOException is thrown
	 * @throws CheckedExceptionSuppressorException
	 *             if CannotCompileException is thrown
	 */
	public static byte[] toBytecode(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		try {
			return ctClass.toBytecode();
		}
		catch (IOException e) {
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
		catch (CannotCompileException e) {
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link CtBehavior#getParameterTypes()}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link CheckedExceptionSuppressorException}.
	 * 
	 * @param ctBehavior
	 *            target {@link CtBehavior}
	 * @return return value of {@link CtBehavior#getParameterTypes()}
	 * @throws IllegalArgumentException
	 *             if <code>ctBehavior</code> is null
	 * @throws CheckedExceptionSuppressorException
	 *             if NotFoundException is thrown
	 */
	public static CtClass[] getParameterTypes(CtBehavior ctBehavior) {
		if (ctBehavior == null) throw new IllegalArgumentException("ctBehavior is null.");
		try {
			return ctBehavior.getParameterTypes();
		}
		catch (NotFoundException e) {
			// Mock test is impossible because CtMethod is final class.
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link CtMethod#getReturnType()}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link CheckedExceptionSuppressorException}.
	 * 
	 * @param ctMethod
	 *            target {@link CtMethod}
	 * @return return value of {@link CtMethod#getReturnType()}
	 * @throws IllegalArgumentException
	 *             if <code>ctMethod</code> is null
	 * @throws CheckedExceptionSuppressorException
	 *             if NotFoundException is thrown
	 */
	public static CtClass getReturnType(CtMethod ctMethod) {
		if (ctMethod == null) throw new IllegalArgumentException("ctMethod is null.");
		try {
			return ctMethod.getReturnType();
		}
		catch (NotFoundException e) {
			// Mock test is impossible because CtMethod is final class.
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link CtClass#getSuperclass()}. Interface's super class is java.lang.Object.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link CheckedExceptionSuppressorException}.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return return value of {@link CtClass#getSuperclass()}
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 * @throws CheckedExceptionSuppressorException
	 *             if NotFoundException is thrown
	 */
	public static CtClass getSuperClass(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		try {
			return ctClass.getSuperclass();
		}
		catch (NotFoundException e) {
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link CtClass#getInterfaces()}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link CheckedExceptionSuppressorException}.
	 * 
	 * @see #getAssignableInterfacesExceptItself(CtClass)
	 * @param ctClass
	 *            target {@link CtClass}
	 * @return return value of {@link CtClass#getInterfaces()}
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 * @throws CheckedExceptionSuppressorException
	 *             if NotFoundException is thrown
	 */
	public static CtClass[] getInterfaces(CtClass ctClass) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		try {
			return ctClass.getInterfaces();
		}
		catch (NotFoundException e) {
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}
	
	// Checked(9.11.2009), Tested(9.11.2009)
	/**
	 * Checked exception suppressor for {@link ClassPool#appendClassPath(String)}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link CheckedExceptionSuppressorException}.
	 * 
	 * @param classPool
	 *            target {@link ClassPool}
	 * @param classpath
	 *            target classpath
	 * @return return value of {@link ClassPool#appendClassPath(String)}
	 * @throws IllegalArgumentException
	 *             if <code>classPool</code> is null
	 * @throws IllegalArgumentException
	 *             if <code>classpath</code> is null
	 * @throws CheckedExceptionSuppressorException
	 *             if NotFoundException is thrown
	 */
	public static ClassPath appendClassPath(ClassPool classPool, String classpath) {
		if (classPool == null) throw new IllegalArgumentException("classPool is null.");
		if (classpath == null) throw new IllegalArgumentException("classpath is null.");
		try {
			return classPool.appendClassPath(classpath);
		}
		catch (NotFoundException e) {
			// NotFoundException is not thrown is javassist source code in 3.11.0.GA.
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}
	
	// Checked(9.10.2009), Tested(9.14.2009)
	/**
	 * Checked exception suppressor for {@link CtClass#getDeclaredMethod(String, CtClass[])}.
	 * </p>
	 * This method converts {@link NotFoundException} to {@link CheckedExceptionSuppressorException}.
	 * 
	 * @param ctClass
	 *            target {@link CtClass}
	 * @param methodName
	 *            target method name
	 * @param parameterTypes
	 *            target method parameter types
	 * @return return value of {@link CtClass#getDeclaredMethod(String, CtClass[])}
	 * @throws IllegalArgumentException
	 *             if <code>ctClass</code> is null
	 * @throws IllegalArgumentException
	 *             if <code>methodName</code> is null
	 * @throws IllegalArgumentException
	 *             if <code>parameterTypes</code> is null
	 * @throws CheckedExceptionSuppressorException
	 *             if NotFoundException is thrown
	 */
	public static CtMethod getDeclaredMethod(CtClass ctClass, String methodName, CtClass[] parameterTypes) {
		if (ctClass == null) throw new IllegalArgumentException("ctClass is null.");
		if (methodName == null) throw new IllegalArgumentException("methodName is null.");
		if (parameterTypes == null) throw new IllegalArgumentException("parameterTypes is null.");
		try {
			return ctClass.getDeclaredMethod(methodName, parameterTypes);
		}
		catch (NotFoundException e) {
			throw new CheckedExceptionSuppressorException(e.getMessage(), e);
		}
	}
	
	/**
	 * Thrown if checked exception is thrown in checked exception suppressor 
	 */
	public static class CheckedExceptionSuppressorException extends IllegalStateException {
		CheckedExceptionSuppressorException(String message, Throwable cause) { super(message, cause); }
	}
}
